""" 
 * ====================================================================
 * Copyright (c) 2001 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *      the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
"""

import mod_snake
from mod_snake import OK, DECLINED, TAKE1, FLAG, OR_AUTHCFG, RSRC_CONF
from mod_snake import APLOG_ERR, HTTP_UNAUTHORIZED, HTTP_INTERNAL_SERVER_ERROR
from mod_snake import RAW_ARGS
from mod_snake_util import set_dir_flag
import string, sys, re, threading

try:    from cPickle import dumps
except: from pickle  import dumps

MOD_NAME     = 'AuthDBAPI'
MOD_VERSION  = '0.1'

class AuthDBAPI:
    def __init__(self, module):
        module.add_hook('check_user_id',     self.check_user_id)
        module.add_hook('auth_checker',      self.auth_checker)
        module.add_hook('create_svr_config', self.create_svr_config)
        module.add_hook('create_dir_config', self.create_dir_config)
	directives = {
            "AuthDBAPI_Authoritative":(OR_AUTHCFG,FLAG,set_dir_flag('auth')),
            "AuthDBAPI_UseDB" :       (OR_AUTHCFG,RAW_ARGS,self.cmd_UseDB),
            "AuthDBAPI_DBLayout" :    (OR_AUTHCFG,TAKE1,self.cmd_DBLayout),
            "AuthDBAPI_PasswdFormat" :(OR_AUTHCFG,TAKE1,self.cmd_PasswdFormat),
            "AuthDBAPI_Driver" :      (RSRC_CONF, TAKE1,self.cmd_Driver),
	}
	module.add_directives(directives)
        self.pword_checkers = { 'none'   : pword_check_none,
                                'apache' : pword_check_apache, }

        module.add_version_component('%s/%s' % (MOD_NAME, MOD_VERSION))
            
    def cmd_DBLayout(self, per_dir, per_svr, layout):
        layout = string.split(layout, ':')
        if len(layout) != 4:
            return 'a table layout: table:namefield:pwordfield:groupfield'
        if not len(layout[0]):
            return 'the table field is required'
        if not len(layout[1]):
            return 'the name field is required'
        per_dir['layout'] = layout

    def cmd_PasswdFormat(self, per_dir, per_svr, arg):
        if arg not in self.pword_checkers.keys():
            return 'a password format in: %s ' % (self.pword_checkers.keys(),)
        per_dir['pwformat'] = arg
    
    def cmd_UseDB(self, per_dir, per_svr, arg):
        try:
            kwargs = eval(arg, {}, {})
            if type(kwargs) != type({}): raise TypeError
        except:
            return '"%s" UseDB: Error parsing argument.  It must be a Python dict: i.e. {"driver" : "gadfly", "databasename" : "mydbase", "directory" : "/dbases"}' % arg
        if not kwargs.has_key('driver'):
            return "UseDB:  A driver is required.  I.e. { 'driver' : 'gadfly', ... }"
 
        if kwargs['driver'] not in per_svr['drivers'].keys():
            return "invalid driver, %s" % kwargs['driver']
        per_dir['usedb'] = kwargs
        
    def cmd_Driver(self, per_dir, per_svr, driver):
        try:
            module = __import__(driver, {},{}, [])
        except:
            nfo = sys.exc_info()
            return 'AuthDBAPI: Error loading driver %s, %s' % (
                driver, nfo[1] or sys.exc_info()[0])

        moddata = { 'apilevel'     : '1.0',
                    'mod_lock'     : threading.Lock(),  # lock of the module
                    'threadsafety' : 0,                 
                    'conncache'    : {}, # Cache of connections
                    }
        if hasattr(module, 'apilevel'):
            # DBAPI 2.0 module
            moddata['apilevel'] = module.apilevel
            moddata['threadsafety'] = module.threadsafety
            moddata['cfunc']      = module.__dict__['connect']
        else:
            moddata['cfunc']      = module.__dict__[driver]
            
        per_svr['drivers'][driver] = moddata

    def create_svr_config(self, server):
        return { 'drivers'     : {},     # Information about loaded drivers
                 'server'      : server }

    def create_dir_config(self, path):
        return { 'auth'     : 1,       # Default to Authoritative
                 'layout'   : None,
                 'usedb'    : None,
                 'pwformat' : 'apache', }

    def return_auth_fail(self, per_dir, r, msg, force):
        if per_dir['auth'] or force:
            mod_snake.ap_log_error(APLOG_ERR, r.server, msg)
            r.note_basic_auth_failure()
            return HTTP_UNAUTHORIZED
        return OK

    def perform_db_query(self, per_svr, usedb, layout, user):
        driver  = usedb['driver']
        drvinfo = per_svr['drivers'][driver]
        cfunc = drvinfo['cfunc']
        usedb = usedb.copy()
        del usedb['driver']

        table, userfield, pwordfield, groupfield = layout
        res = { 'user' : None, 'pword' : None, 'group' : None }
        selfields = [(userfield, 'user')]
        if len(pwordfield): selfields.append((pwordfield, 'pword'))
        if len(groupfield): selfields.append((groupfield, 'group'))
        selstring = string.join(map(lambda (x,y): x, selfields), ',')
        querystr = "SELECT %s FROM %s WHERE %s='%s'" % (
            selstring, table, userfield, user)

        # Threadsafety < 1 check
        if not drvinfo['threadsafety']: drvinfo['mod_lock'].acquire()
        # Look for a cached connection -- cheezy cache-mechanism
        useddb_string = dumps(usedb)
        conn_lock = None
        if drvinfo['conncache'].has_key(useddb_string):
            conn, conn_lock = drvinfo['conncache'][useddb_string]
            if drvinfo['threadsafety'] < 2:
                sys.stderr.write('Locking ... ')
                conn_lock.acquire()
                sys.stderr.write('Locked\n')
        else:
            conn = apply(cfunc, (), usedb)
        curs = conn.cursor()
        curs.execute(querystr)
        try:
            try:
                fetched = curs.fetchone()
            except:
                return None
        finally:
            curs.close()

            if not drvinfo['conncache'].has_key(useddb_string):
                drvinfo['conncache'][useddb_string] = conn, threading.Lock()
            else:
                if drvinfo['threadsafety'] < 2:
                    sys.stderr.write('Releasing ... ')
                    conn_lock.release()
                    sys.stderr.write('Released\n')
            if not drvinfo['threadsafety']: drvinfo['mod_lock'].release()

        if not fetched: return None
        for i in range(len(selfields)):
            field, tag = selfields[i]
            res[tag] = fetched[i]
            
        return res

    def get_usedb_and_layout(self, per_dir, per_svr):
        usedb, layout = per_dir['usedb'], per_dir['layout']
        if not usedb or not layout:
            if not usedb:
                mod_snake.ap_log_error(APLOG_ERR, per_svr['server'],
                                    'AuthDBAPI: AuthDBAPI_UseDB not used')
            else:
                mod_snake.ap_log_error(APLOG_ERR, per_svr['server'],
                                    'AuthDBAPI: AuthDBAPI_DBLayout not used')
            raise "AuthDBAPI: Config error occurred"
        return usedb, layout
    
    def check_user_id(self, per_dir, per_svr, r):
        pwres, given_pword = r.get_basic_auth_pw()
        if pwres: return pwres

        usedb, layout = self.get_usedb_and_layout(per_dir, per_svr)
        if not len(layout[2]):  # No password field, thus no authentication
            return DECLINED

        userinfo = self.perform_db_query(per_svr, usedb, layout, r.user)
        if not userinfo:  # Couldn't find the user?
            return self.return_auth_fail(per_dir, r,
                                         "couldn't find user: %s" % r.user,
                                         0)
        pword_checker = self.pword_checkers[per_dir['pwformat']]
        if (r.user == userinfo['user'] and
            pword_checker(given_pword,userinfo['pword'])):
            return OK
        return self.return_auth_fail(per_dir, r,
                                     "user %s: bad pword entered" % r.user,
                                     1)

    def auth_checker(self, per_dir, per_svr, r):
        usedb, layout = self.get_usedb_and_layout(per_dir, per_svr)

        requires = r.requires()
        if not len(layout[3]) or not requires: # No group field, no auth
            return DECLINED
        for require in requires:
            if not (require.method_mask & (1 << r.method_number)): continue
            args = string.split(require.requirement)
            if args[0] != 'group': continue
            userinfo = self.perform_db_query(per_svr, usedb, layout, r.user)
            if not userinfo:  # No user?
                return self.return_auth_fail(per_dir, r,
                                             "couldn't find user: %s" % r.user,
                                             0)
            groups = string.split(userinfo['group'], ',')
            for group in map(lambda x: string.split(x), groups):
                if group[0] in args[1:]: return OK

            return self.return_auth_fail(per_dir, r,
                                         "User %s not in right group" % r.user,
                                         1)
        return DECLINED


def pword_check_apache(pword, check_against):
    return mod_snake.ap_validate_password(pword, check_against)

def pword_check_none(pword, check_against):
    return pword == check_against
